from typing import Any
from instaui.internal.ui.element import Element


def _create_html_element(tag: str):
    class HtmlElement(Element):
        def __init__(self, inner_text: str | Any | None = None, **kwargs):
            super().__init__(tag)
            props = {"innerText": inner_text}
            props.update(kwargs)
            self.props(props)

    HtmlElement.__name__ = tag.capitalize()
    return HtmlElement


a = _create_html_element("a")
abbr = _create_html_element("abbr")
address = _create_html_element("address")
area = _create_html_element("area")
article = _create_html_element("article")
aside = _create_html_element("aside")
audio = _create_html_element("audio")
b = _create_html_element("b")
bdi = _create_html_element("bdi")
bdo = _create_html_element("bdo")
blockquote = _create_html_element("blockquote")
br = _create_html_element("br")
canvas = _create_html_element("canvas")
caption = _create_html_element("caption")
cite = _create_html_element("cite")
code = _create_html_element("code")
col = _create_html_element("col")
colgroup = _create_html_element("colgroup")
data = _create_html_element("data")
datalist = _create_html_element("datalist")
dd = _create_html_element("dd")
del_ = _create_html_element("del")
details = _create_html_element("details")
dfn = _create_html_element("dfn")
div = _create_html_element("div")
dl = _create_html_element("dl")
dt = _create_html_element("dt")
em = _create_html_element("em")
embed = _create_html_element("embed")
fieldset = _create_html_element("fieldset")
figcaption = _create_html_element("figcaption")
figure = _create_html_element("figure")
footer = _create_html_element("footer")
form = _create_html_element("form")
h1 = _create_html_element("h1")
h2 = _create_html_element("h2")
h3 = _create_html_element("h3")
h4 = _create_html_element("h4")
h5 = _create_html_element("h5")
h6 = _create_html_element("h6")
header = _create_html_element("header")
hr = _create_html_element("hr")
i = _create_html_element("i")
iframe = _create_html_element("iframe")
ins = _create_html_element("ins")
kbd = _create_html_element("kbd")
label = _create_html_element("label")
legend = _create_html_element("legend")
li = _create_html_element("li")
main = _create_html_element("main")
map_ = _create_html_element("map")
mark = _create_html_element("mark")
menu = _create_html_element("menu")
meta = _create_html_element("meta")
meter = _create_html_element("meter")
nav = _create_html_element("nav")
noscript = _create_html_element("noscript")
object_ = _create_html_element("object")
ol = _create_html_element("ol")
optgroup = _create_html_element("optgroup")
option = _create_html_element("option")
output = _create_html_element("output")
p = _create_html_element("p")
param = _create_html_element("param")
picture = _create_html_element("picture")
pre = _create_html_element("pre")
progress = _create_html_element("progress")
q = _create_html_element("q")
rp = _create_html_element("rp")
rt = _create_html_element("rt")
ruby = _create_html_element("ruby")
s = _create_html_element("s")
samp = _create_html_element("samp")
script = _create_html_element("script")
search = _create_html_element("search")
section = _create_html_element("section")
small = _create_html_element("small")
source = _create_html_element("source")
span = _create_html_element("span")
strong = _create_html_element("strong")
style = _create_html_element("style")
sub = _create_html_element("sub")
summary = _create_html_element("summary")
sup = _create_html_element("sup")
svg = _create_html_element("svg")
tbody = _create_html_element("tbody")
td = _create_html_element("td")
template = _create_html_element("template")
tfoot = _create_html_element("tfoot")
th = _create_html_element("th")
thead = _create_html_element("thead")
time = _create_html_element("time")
title = _create_html_element("title")
tr = _create_html_element("tr")
track = _create_html_element("track")
u = _create_html_element("u")
var = _create_html_element("var")
wbr = _create_html_element("wbr")



import inspect
from instaui.internal.backend.common.response import response_data
from .watch_handler import get_watch_handler
from .event_handler import get_event_handler
from .async_executor import AsyncExecutor


class Api:
    def __init__(self) -> None:
        self._async_executor = AsyncExecutor()

    def watch_call(self, data: dict):
        hkey = data.pop("hKey")
        handler = get_watch_handler(hkey)
        if handler is None:
            return {"error": "watch handler not found"}

        exec_result = handler.fn(*handler.get_handler_args(_get_binds_from_data(data)))

        if inspect.iscoroutinefunction(handler.fn):
            result = self._async_executor.run(exec_result)
        else:
            result = exec_result
        return response_data(handler.outputs_binding_count, result)

    def event_call(self, data: dict):
        hkey = data.pop("hKey")
        handler = get_event_handler(hkey)
        if handler is None:
            raise ValueError("event handler not found")

        args = [bind for bind in data.get("bind", [])]

        exec_result = handler.fn(*handler.get_handler_args(args))

        if inspect.iscoroutinefunction(handler.fn):
            result = self._async_executor.run(exec_result)
        else:
            result = exec_result

        return response_data(handler.outputs_binding_count, result)


def _get_binds_from_data(data: dict):
    return data.get("input", [])



from typing import Optional
from instaui.internal.pipeline.utils import html_utils
from instaui.systems.dataclass_system import dataclass, field


@dataclass()
class HtmlRenderContext:
    app_var: str
    lang: str = "en"
    title: str = ""
    favicon_url: str = ""

    css_links: list[str] = field(default_factory=list)
    style_tags: list[tuple[Optional[str], str, dict]] = field(default_factory=list)
    """(group_id, content, attrs)"""

    script_tags_in_head: list[tuple[Optional[dict], str | None]] = field(
        default_factory=list
    )
    """(attrs_dict, content)"""
    script_tags_in_body: list[tuple[Optional[dict], str | None]] = field(
        default_factory=list
    )
    """(attrs_dict, content)"""
    entry_script_type: str = "module"
    entry_code: str = ""

    mount_id: str = "app"

    import_map: Optional[dict[str, str]] = None

    icons_svg_fragment: Optional[str] = None

    def render_attrs(self, attrs: Optional[dict]) -> str:
        return html_utils.render_attrs(attrs)



from typing import Any
from instaui.internal.ui.element import Element
from instaui.internal.ui.value_element import ValueElement
from instaui.internal.ui.disabled_element import DisabledElement
from ._mixins import InputEventMixin


class Input(InputEventMixin, DisabledElement, ValueElement[str]):
    def __init__(
        self,
        value: str | Any | None = None,
        *,
        model_value: str | Any | None = None,
        disabled: bool | None = None,
        type: str | None = None,
        **kwargs,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props({"type": type, "disabled": disabled, "value": model_value, **kwargs})

    def _input_event_mixin_element(self) -> Element:
        return self



from typing import Any, Mapping, Optional, Type, TypeVar, cast, overload
from instaui.debug.api_boundary import user_api
from instaui.debug.source import get_source_span
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.bindable import BindableMixin, VarableBindHelper
from instaui.internal.ui.enums import InputBindingType
from instaui.internal.ui.path_var import PathVar
from instaui.internal.ui.protocol import ObservableProtocol
from instaui.internal.ui.variable import Variable

_R_TYPE = TypeVar("_R_TYPE")


class ExprComputed(Variable, PathVar, BindableMixin, ObservableProtocol):
    def __init__(
        self,
        code: str,
        bindings: Optional[Mapping[str, Any]] = None,
    ) -> None:
        self._source_span_ = get_source_span()
        new_scope_if_needed()
        define_scope = get_current_scope()
        self.code = code
        self._bindings = UILiteralExpr.try_parse_dict(bindings) if bindings else None

        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_expr_computed,
            define_scope=define_scope,
            lazy_mark_used=[self._bindings],
        )

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref


TExprComputed = ExprComputed


@overload
def expr_computed(
    code: str,
    *,
    bindings: Optional[Mapping[str, Any]] = None,
) -> Any: ...


@overload
def expr_computed(
    code: str,
    *,
    bindings: Optional[Mapping[str, Any]] = None,
    r_type: Optional[Type[_R_TYPE]] = None,
) -> _R_TYPE: ...


@user_api
def expr_computed(
    code: str,
    *,
    bindings: Optional[Mapping[str, Any]] = None,
    r_type: Optional[Type[_R_TYPE]] = None,
) -> Any | _R_TYPE:
    """
    A client-side computed property that evaluates JavaScript code to derive reactive values.

    Args:
        code (str): JavaScript code to evaluate.
        bindings (Optional[Mapping[str, Any]], optional): Variables to bind to the code. Defaults to None.
        r_type (Optional[Type[_R_TYPE]], optional): Type hint for the return value. Defaults to None.


    """

    computed = ExprComputed(
        code=code,
        bindings=bindings,
    )

    if r_type is None:
        return computed

    return cast(_R_TYPE, computed)



from __future__ import annotations
from typing import Sequence
from instaui.internal.ui.element import Element
from .simple import li
from instaui.internal.ui.vfor import VFor


class Ul(Element):
    def __init__(self):
        super().__init__("ul")

    @classmethod
    def from_list(cls, data: Sequence) -> Ul:
        with Ul() as ul:
            with VFor(data) as items:
                li(items)

        return ul



from dataclasses import dataclass
import inspect
import types
from typing import Callable, List, Any


@dataclass
class DefaultParam:
    name: str
    default: Any


def extract_defaults_and_strip(func: Callable):
    sig = inspect.signature(func)
    defaults: List[DefaultParam] = []
    for p in sig.parameters.values():
        if p.default is not inspect._empty:
            defaults.append(DefaultParam(p.name, p.default))

    if not defaults:
        return [], func

    # new function with no positional defaults
    new_func = types.FunctionType(
        func.__code__,
        func.__globals__,
        name=func.__name__,
        argdefs=None,  # key point：remove positional defaults
        closure=func.__closure__,
    )

    # remove keyword-only defaults
    new_func.__kwdefaults__ = None

    new_func.__dict__.update(func.__dict__)
    new_func.__annotations__ = getattr(func, "__annotations__", {}).copy()
    new_func.__doc__ = func.__doc__
    new_func.__module__ = func.__module__

    return defaults, new_func



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional
from instaui.systems.dataclass_system import dataclass

if TYPE_CHECKING:
    from .plugin import PluginDependencyInfo
    from instaui.internal.assets.style_assets import StyleTag


@dataclass(frozen=True)
class AssetsSnapshot:
    css_links: list[str | Path]
    style_tags: list[StyleTag]
    script_tags_in_head: list[ScriptTagSnapshot]
    script_tags_in_body: list[ScriptTagSnapshot]
    import_maps: dict[str, str | Path]
    favicon: Optional[str | Path]
    plugins: set[PluginDependencyInfo]
    component_dependencies: list[ComponentDependencySnapshot]


@dataclass(frozen=True)
class ComponentDependencySnapshot:
    tag: str
    esm: Path


@dataclass(frozen=True)
class ScriptTagSnapshot:
    inline_code: str | None = None
    attrs: dict[str, Any] | None = None



__all__ = [
    "button",
    "checkbox",
    "date",
    "dialog",
    "image",
    "img",
    "input",
    "number",
    "radio",
    "range",
    "select",
    "a",
    "abbr",
    "address",
    "area",
    "article",
    "aside",
    "audio",
    "b",
    "bdi",
    "bdo",
    "blockquote",
    "br",
    "canvas",
    "caption",
    "cite",
    "code",
    "col",
    "colgroup",
    "data",
    "datalist",
    "dd",
    "del_",
    "details",
    "dfn",
    "div",
    "dl",
    "dt",
    "em",
    "embed",
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "header",
    "hr",
    "i",
    "iframe",
    "ins",
    "kbd",
    "label",
    "legend",
    "li",
    "main",
    "map_",
    "mark",
    "menu",
    "meta",
    "meter",
    "nav",
    "noscript",
    "object_",
    "ol",
    "optgroup",
    "option",
    "output",
    "p",
    "param",
    "picture",
    "pre",
    "progress",
    "q",
    "rp",
    "rt",
    "ruby",
    "s",
    "samp",
    "script",
    "search",
    "section",
    "small",
    "source",
    "span",
    "strong",
    "style",
    "sub",
    "summary",
    "sup",
    "svg",
    "tbody",
    "td",
    "template",
    "tfoot",
    "th",
    "thead",
    "time",
    "title",
    "tr",
    "track",
    "u",
    "var",
    "wbr",
    "paragraph",
    "table",
    "textarea",
    "ul",
    "video",
]

from .button import Button as button
from .checkbox import Checkbox as checkbox
from .date import Date as date
from .dialog import Dialog as dialog
from .image import Image as image
from .image import Image as img
from .input import Input as input
from .number import Number as number
from .radio import Radio as radio
from .range import Range as range
from .select import Select as select
from .simple import (
    a,
    abbr,
    address,
    area,
    article,
    aside,
    audio,
    b,
    bdi,
    bdo,
    blockquote,
    br,
    canvas,
    caption,
    cite,
    code,
    col,
    colgroup,
    data,
    datalist,
    dd,
    del_,
    details,
    dfn,
    div,
    dl,
    dt,
    em,
    embed,
    fieldset,
    figcaption,
    figure,
    footer,
    form,
    h1,
    h2,
    h3,
    h4,
    h5,
    h6,
    header,
    hr,
    i,
    iframe,
    ins,
    kbd,
    label,
    legend,
    li,
    main,
    map_,
    mark,
    menu,
    meta,
    meter,
    nav,
    noscript,
    object_,
    ol,
    optgroup,
    option,
    output,
    p,
    param,
    picture,
    pre,
    progress,
    q,
    rp,
    rt,
    ruby,
    s,
    samp,
    script,
    search,
    section,
    small,
    source,
    span,
    strong,
    style,
    sub,
    summary,
    sup,
    svg,
    tbody,
    td,
    template,
    tfoot,
    th,
    thead,
    time,
    title,
    tr,
    track,
    u,
    var,
    wbr,
)
from .simple import (
    p as paragraph,
)
from .table import Table as table
from .textarea import Textarea as textarea
from .ul import Ul as ul
from .video import Video as video



from typing import Any, Literal
from instaui.internal.ui.element import Element


TText = str | Any | None


class Heading(Element):
    def __init__(
        self,
        inner_text: TText = None,
        *,
        level: Literal[1, 2, 3, 4, 5, 6] = 1,
    ):
        super().__init__(f"h{level}")
        self.props(
            {
                "innerText": inner_text,
            }
        )



from typing import TypeVar, overload, Any


T = TypeVar("T")
WEB_EVENT_PARAM = object()


@overload
def event_param() -> Any: ...


@overload
def event_param(type_: type[T]) -> T: ...


def event_param(type_: type | None = None):
    """
    Create a placeholder for an event-bound parameter in a UI event handler.

    This function returns a special marker that can be used as the default value
    for a handler parameter in `@ui.event` functions. The actual value of the
    parameter will be injected dynamically when the event is triggered, typically
    via the `params` argument of `component.on()`.

    Args:
        type_ (type | None, optional): Optional Python type hint for the parameter. This can be used for
            type checking or documentation purposes but does not affect runtime
            behavior.

    Example:
    .. code-block:: python

        @ui.event
        def on_click(index = ui.event_param(int)):
            print(f"Clicked on item {index}")

        html.button('Click me').on_click(on_click, params=[42])
    """
    return WEB_EVENT_PARAM


def is_event_param(obj):
    return obj is WEB_EVENT_PARAM



from contextvars import copy_context
from typing import Callable, TypeVar

T = TypeVar("T")


class PageExecutionScope:
    """
    Defines the execution boundary of a single page rendering.
    Responsible for isolating ContextVar-based states.
    """

    @staticmethod
    def run(fn: Callable[[], T]) -> T:
        ctx = copy_context()
        return ctx.run(fn)



from instaui.internal.ui.element import Element
from instaui.internal.ui.value_element import ValueElement
from ._mixins import InputEventMixin


_T_value = int | float | None


class Number(InputEventMixin, ValueElement[_T_value]):
    def __init__(
        self,
        value: _T_value = None,
        *,
        model_value: _T_value = None,
        min: _T_value = None,
        max: _T_value = None,
        **kwargs,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props(
            {"type": "number", "min": min, "max": max, "value": model_value, **kwargs}
        )

    def _input_event_mixin_element(self) -> Element:
        return self



from typing import Any
from instaui.internal.ui.element import Element
from instaui.internal.ui.value_element import ValueElement
from ._mixins import InputEventMixin


class Checkbox(InputEventMixin, ValueElement[bool | str | list]):
    def __init__(
        self,
        value: bool | str | list | None = None,
        *,
        model_value: bool | str | None = None,
        checked: bool | None = None,
        id: Any = None,
        **kwargs,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props(
            {
                "type": "checkbox",
                "id": id,
                "checked": checked,
                "value": model_value,
                **kwargs,
            }
        )

    def _input_event_mixin_element(self) -> Element:
        return self



from typing import Callable, Generic, Hashable, Optional, Sequence, TypeVar, overload
from typing_extensions import ParamSpec
from instaui.debug.api_boundary import user_api
from instaui.debug.model import SourceSpan
from instaui.debug.source import get_source_span
from instaui.internal.codec import build_type_adapter_map
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.event_context import DatasetEventContext
from instaui.internal.ui.pre_setup import normalize_pre_setup
from instaui.internal.ui.utils.func_utils import extract_defaults_and_strip
from instaui.protocol.invocation.spec import EventSpec

from .variable import Variable
from .event_modifier import TEventModifier
from .app_context import get_app, get_current_scope
from .event import EventMixin
from .utils.validators import ensure_output_list
from .bindable import BindableMixin, VarableBindHelper, mark_used
from .web_event_param import is_event_param

P = ParamSpec("P")
R = TypeVar("R")


class WebEvent(EventMixin, BindableMixin, Variable, Generic[P, R]):
    def __init__(
        self,
        fn: Callable[P, R],
        inputs: Sequence,
        outputs: Sequence,
        pre_setup: list | None = None,
        modifier: Sequence[TEventModifier] | None = None,
        *,
        extra_key: Optional[Sequence[Hashable]] = None,
        _source_span_: SourceSpan,
    ):
        self._source_span_ = _source_span_
        new_scope_if_needed()
        ensure_output_list(outputs)
        define_scope = get_current_scope()

        if not inputs:
            default_params, fn = extract_defaults_and_strip(fn)
            inputs = [x.default for x in default_params]

        inputs = [input for input in inputs if not is_event_param(input)]

        self._inputs = UILiteralExpr.try_parse_list(inputs)
        self._outputs = outputs
        self._fn = fn
        self._pre_setup = normalize_pre_setup(pre_setup or [])
        self._modifier = modifier or []
        self._extra_key = extra_key

        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_web_event,
            define_scope=define_scope,
            lazy_mark_used=[
                *self._inputs,
                *self._outputs,
                *[x[0] for x in self._pre_setup],
            ],
        )

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self._fn(*args, **kwargs)

    def _attach_to_element(
        self,
        *,
        params: Sequence | None = None,
        modifier: Sequence[TEventModifier] | None = None,
    ):
        mark_used(params)
        mark_used(self)

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    def _export_invocation_spec(self) -> EventSpec:
        type_adapter_map = build_type_adapter_map(self._fn, self._inputs)
        dataset_input_indexs = [
            i
            for i, input_ in enumerate(self._inputs or [])
            if isinstance(input_, DatasetEventContext)
        ]
        return EventSpec(
            len(self._outputs or []),
            dataset_input_indexs,
            type_adapter_map,
            extra_key=self._extra_key,
        )


@overload
def event(_func: Callable[P, R]) -> WebEvent[P, R]: ...


@overload
def event(
    *,
    inputs: Sequence | None = ...,
    outputs: Sequence | None = ...,
    pre_setup: list | None = ...,
    extra_key: Sequence[Hashable] | None = ...,
) -> Callable[[Callable[P, R]], WebEvent[P, R]]: ...


@user_api
def event(
    _func: Callable[P, R] | None = None,
    *,
    inputs: Sequence | None = None,
    outputs: Sequence | None = None,
    pre_setup: list | None = None,
    extra_key: Sequence[Hashable] | None = None,
) -> WebEvent[P, R] | Callable[[Callable[P, R]], WebEvent[P, R]]:
    """
    Creates an event handler decorator for binding reactive logic to component events.

    Args:
        inputs (Optional[Sequence], optional): Reactive sources (state objects, computed properties)
                                   that should be accessible during event handling.
                                   These values will be passed to the decorated function
                                   when the event fires.
        outputs (Optional[Sequence], optional): Targets (state variables, UI elements) that should
                                    update when this handler executes. Used for coordinating
                                    interface updates after the event is processed.
        pre_setup (Optional[list], optional): A list of pre-setup actions to be executed before the event executes.


    # Example:
    .. code-block:: python
        from instaui import ui, html

        a = ui.state(0)

        @ui.event(inputs=[a], outputs=[a])
        def plus_one(a):
            return a + 1

        html.button("click me").on_click(plus_one)
        html.paragraph(a)

    use pre_setup:
    .. code-block:: python
        a = ui.state(0)
        task_running = ui.state(False)

        @ui.event(inputs=[a], outputs=[a], pre_setup=[task_running,True,False])
        async def long_running_task(a):
            await asyncio.sleep(3)
            return a + 1

        html.button("click me").on_click(long_running_task).disabled(task_running)
    """

    if get_app().mode == "zero":
        raise Exception(
            "Cannot use event decorator in zero mode. You should use `ui.js_event` instead."
        )

    _source_span_ = get_source_span()

    def decorator(func: Callable[P, R]):
        return WebEvent(
            func,
            inputs or [],
            outputs=outputs or [],
            pre_setup=pre_setup,
            extra_key=extra_key,
            _source_span_=_source_span_,
        )

    # case：@event
    if _func is not None and callable(_func):
        return decorator(_func)

    # case：@event(...)
    return decorator



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Protocol
import logging

if TYPE_CHECKING:
    from instaui.internal.ui.element import Element

logger = logging.getLogger(__name__)


class ElementCreateHook(Protocol):
    priority: int

    def on_element_create(self, element: Element) -> None: ...


@dataclass(order=True)
class _HookEntry:
    priority: int
    hook: ElementCreateHook = field(compare=False)


_element_create_hooks: List[_HookEntry] = []


# -------------------------
# Public API
# -------------------------


def register_element_create_hook(hook: ElementCreateHook) -> None:
    """Register a hook. Higher priority executes first."""
    priority = getattr(hook, "priority", 0)
    entry = _HookEntry(priority=priority, hook=hook)
    _element_create_hooks.append(entry)
    _element_create_hooks.sort(reverse=True)


def clear_element_create_hooks() -> None:
    """Clear all registered hooks."""
    _element_create_hooks.clear()


# -------------------------
# Internal
# -------------------------


def trigger_element_create(element: Element) -> None:
    for entry in _element_create_hooks:
        try:
            entry.hook.on_element_create(element)
        except Exception:
            logger.exception(
                "ElementCreateHook failed: %s", entry.hook.__class__.__name__
            )



from instaui.internal.ui.element import Element
from instaui.internal.ui.event import EventMixin
from instaui.internal.ui.event_modifier import TEventModifier


class Dialog(Element):
    """HTML dialog element for creating modal or non-modal dialogs.

    Corresponds to the HTML <dialog> tag, supporting dialog show, close, and cancel events.
    The initial visibility of the dialog can be controlled through the open property.

    Args:
        title (str | None, optional): Dialog title. Defaults to None.
        open (bool, optional): Whether the dialog is initially open. Defaults to False.
        on_close (EventMixin | None, optional): Event handler for dialog close event. Defaults to None.
        on_cancel (EventMixin | None, optional): Event handler for dialog cancel event (triggered by ESC key). Defaults to None.
        on_show (EventMixin | None, optional): Event handler for dialog show event. Defaults to None.
    """

    def __init__(
        self,
        *,
        title: str | None = None,
        open: bool = False,
        on_close: EventMixin | None = None,
        on_cancel: EventMixin | None = None,
        on_show: EventMixin | None = None,
        **kwargs,
    ):
        super().__init__("dialog")
        self.props({"title": title, "open": open, **kwargs})

        if on_close:
            self.on_close(on_close)
        if on_cancel:
            self.on_cancel(on_cancel)
        if on_show:
            self.on_show(on_show)

    def on_close(
        self,
        handler: EventMixin,
        *,
        params: list | None = None,
        modifier: list[TEventModifier] | None = None,
    ):
        """Bind dialog close event.

        Triggered when the dialog is closed via the close() method.

        Args:
            handler (EventMixin): Event handler function.
            params (list | None, optional): Parameters to pass to the handler. Defaults to None.
            modifier (list[TEventModifier] | None, optional): Event modifiers. Defaults to None.

        Returns:
            Dialog: Returns self instance for method chaining.
        """
        self.on("close", handler, params=params, modifier=modifier)
        return self

    def on_cancel(
        self,
        handler: EventMixin,
        *,
        params: list | None = None,
        modifier: list[TEventModifier] | None = None,
    ):
        """Bind dialog cancel event.

        Triggered when the user closes the dialog by pressing the ESC key.
        Note: Only modal dialogs opened with showModal() support ESC key cancellation.

        Args:
            handler (EventMixin): Event handler function.
            params (list | None, optional): Parameters to pass to the handler. Defaults to None.
            modifier (list[TEventModifier] | None, optional): Event modifiers. Defaults to None.

        Returns:
            Dialog: Returns self instance for method chaining.
        """
        self.on("cancel", handler, params=params, modifier=modifier)
        return self

    def on_show(
        self,
        handler: EventMixin,
        *,
        params: list | None = None,
        modifier: list[TEventModifier] | None = None,
    ):
        """Bind dialog show event.

        Triggered when the dialog is displayed via show() or showModal() methods.

        Args:
            handler (EventMixin): Event handler function.
            params (list | None, optional): Parameters to pass to the handler. Defaults to None.
            modifier (list[TEventModifier] | None, optional): Event modifiers. Defaults to None.

        Returns:
            Dialog: Returns self instance for method chaining.
        """
        self.on("show", handler, params=params, modifier=modifier)
        return self



from instaui.internal.ui.element import Element
from instaui.internal.ui.event import EventMixin
from instaui.internal.ui.event_modifier import TEventModifier
from instaui.internal.ui.disabled_element import DisabledElement


class Button(Element, DisabledElement):
    def __init__(self, inner_text: str | None = None, **kwargs):
        super().__init__("button")

        self.props(
            {
                "innerText": inner_text,
                **kwargs,
            }
        )

    def on_click(
        self,
        handler: EventMixin,
        *,
        params: list | None = None,
        modifier: list[TEventModifier] | None = None,
    ):
        self.on("click", handler, params=params, modifier=modifier)
        return self



from __future__ import annotations
from typing import TYPE_CHECKING

from instaui.internal.ui.value_element import ValueElement
from ._mixins import InputEventMixin

if TYPE_CHECKING:
    from instaui.internal.ui.element import Element

_T_value = str


class Date(InputEventMixin, ValueElement[_T_value]):
    def __init__(
        self,
        value: _T_value | None = None,
        *,
        model_value: _T_value | None = None,
        **kwargs,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props({"type": "date", "value": model_value, **kwargs})

    def _input_event_mixin_element(self) -> Element:
        return self  # type: ignore



from instaui.internal.ui.input_slient_data import InputSilentData
from instaui.internal.ui.bindable import is_bindable


def auto_made_inputs_to_slient(
    inputs: list,
    outputs: list,
):
    if (not inputs) or (not outputs):
        return inputs

    outputs_set = set(outputs)

    return [
        InputSilentData(input) if is_bindable(input) and input in outputs_set else input
        for input in inputs
    ]



from itertools import chain, groupby
from pathlib import Path
from instaui.internal.assets.base import AssetsDeclaration
from instaui.internal.assets.component_dep import ComponentDependencyRegistry
from instaui.internal.assets.css_assets import CssAssetCollection
from instaui.internal.assets.script_assets import JSAsset
from instaui.internal.assets.snapshot import (
    AssetsSnapshot,
    ComponentDependencySnapshot,
    ScriptTagSnapshot,
)
from instaui.internal.assets.style_assets import StyleTag
from instaui.systems.dict_system import merge_dicts
from .processors.base import AssetsProcessor


class AssetsPipeline:
    def __init__(self, processors: list[AssetsProcessor]):
        self._processors = processors

    def run(
        self, global_assets: AssetsDeclaration, local_assets: AssetsDeclaration
    ) -> AssetsSnapshot:
        merged = self.merge(global_assets, local_assets)

        for processor in self._processors:
            processor.process(merged)

        script_tags_in_head = [
            _convert_script_tag(js_asset)
            for js_asset in merged.js_asset
            if js_asset.position == "head"
        ]

        script_tags_in_body = [
            _convert_script_tag(js_asset)
            for js_asset in merged.js_asset
            if js_asset.position == "body"
        ]

        return AssetsSnapshot(
            css_links=list(merged.css_links.iter_css_links()),
            style_tags=merged.style_tags,
            script_tags_in_head=script_tags_in_head,
            script_tags_in_body=script_tags_in_body,
            import_maps=merged.import_maps,
            favicon=merged.favicon,
            plugins=merged.plugins,
            component_dependencies=_convert_component_snapshot(
                merged.component_dependencies
            ),
        )

    def merge(
        self, global_assets: AssetsDeclaration, local_assets: AssetsDeclaration
    ) -> AssetsDeclaration:
        assert local_assets is not None, "should be called inside a page() context"

        plugins = global_assets.plugins | local_assets.plugins

        component_dependencies = ComponentDependencyRegistry()

        for record in chain(
            global_assets.component_dependencies.records,
            local_assets.component_dependencies.records,
        ):
            component_dependencies.add(record.component, record.dependency)

        css_links = CssAssetCollection.merge(
            global_assets.css_links, local_assets.css_links
        )
        style_tags = _merged_style_tags(
            global_assets.style_tags, local_assets.style_tags
        )
        script_tags = global_assets.js_asset + local_assets.js_asset
        import_maps = {**global_assets.import_maps, **local_assets.import_maps}
        favicon = global_assets.favicon or (
            local_assets.favicon if local_assets else None
        )

        component_extensions = global_assets.component_extensions.overridden_by(
            local_assets.component_extensions
        )

        return AssetsDeclaration(
            css_links=css_links,
            style_tags=style_tags,
            js_asset=script_tags,
            import_maps=import_maps,
            favicon=favicon,
            plugins=plugins,
            component_dependencies=component_dependencies,
            component_extensions=component_extensions,
        )


def _merged_style_tags(global_tags: list[StyleTag], local_tags: list[StyleTag]):
    tags = global_tags + local_tags
    sorted_tags = sorted(tags, key=lambda tag: tag.group_id or "")

    def to_style_tag(group_id, group_tags):
        group_tags = tuple(group_tags)

        return StyleTag(
            "\n".join(tag.content for tag in group_tags),
            group_id=group_id,
            attrs=merge_dicts(*[tag.attrs for tag in group_tags]),
        )

    return [
        to_style_tag(group_id, group_tags)
        for group_id, group_tags in groupby(sorted_tags, lambda tag: tag.group_id)
    ]


def _convert_component_snapshot(
    registry: ComponentDependencyRegistry,
) -> list[ComponentDependencySnapshot]:
    return [
        ComponentDependencySnapshot(
            tag=record.dependency.tag_name,
            esm=record.dependency.esm,
        )
        for record in registry.records
    ]


def _convert_script_tag(asset: JSAsset):
    inline_code = (
        _js_source_to_inline_code(asset.source) if asset.kind == "inline" else None
    )

    attrs = dict(asset.attrs)

    if asset.kind != "inline":
        attrs["src"] = asset.source

    if asset.module:
        attrs["type"] = "module"

    if asset.loading == "async":
        attrs["async"] = True
    elif asset.loading == "defer":
        attrs["defer"] = True

    return ScriptTagSnapshot(
        attrs=attrs,
        inline_code=inline_code,
    )


def _js_source_to_inline_code(source: str | Path):
    if isinstance(source, Path):
        return source.read_text(encoding="utf-8")

    return source



from __future__ import annotations

import ast
import inspect
from pathlib import Path
import re
import sys
from typing import (
    Any,
    Callable,
    ClassVar,
    Literal,
    Optional,
    Union,
    cast,
    overload,
    TYPE_CHECKING,
)
from typing_extensions import Self
from collections import defaultdict

from instaui.constants.runtime import RuntimeMode
from instaui.constants.ui import (
    SCOPED_STYLE_GROUP_ID,
    DEFAULT_SLOT_NAME,
    MOUNTED_EVENT_NAME,
)
from instaui.hooks.element_create import trigger_element_create
from instaui.internal.assets import add_component_dependency, ComponentDependencyInfo
from instaui.internal import assets
from instaui.internal.assets.component_dep import ComponentDependencyOverride
from instaui.internal.ui.event import EventMixin
from instaui.internal.ui.event_args import EventArgs
from instaui.internal.ui.event_modifier import TEventModifier
from instaui.internal.ui.slot import Slot
from instaui.internal.ui.variable import Variable
from instaui.systems.json_system import to_json_str
from instaui.types import TVModelModifier
from .bindable import mark_used, is_bindable
from .app_context import get_app, get_current_container, get_current_scope
from .directive import Directive
from .renderable import Renderable


if TYPE_CHECKING:
    from instaui.internal.ui.vfor_item import VForItem
    from instaui.internal.ui.element_ref import ElementRef


TVarGetterStrategy = Union[Literal["as_needed", "all"], list]

# Refer to the NiceGUI project.
# https://github.com/zauberzeug/nicegui/blob/main/nicegui/element.py
PROPS_PATTERN = re.compile(
    r"""
# Match a key-value pair optionally followed by whitespace or end of string
([:\w\-]+)          # Capture group 1: Key
(?:                 # Optional non-capturing group for value
    =               # Match the equal sign
    (?:             # Non-capturing group for value options
        (           # Capture group 2: Value enclosed in double quotes
            "       # Match  double quote
            [^"\\]* # Match any character except quotes or backslashes zero or more times
            (?:\\.[^"\\]*)*  # Match any escaped character followed by any character except quotes or backslashes zero or more times
            "       # Match the closing quote
        )
        |
        (           # Capture group 3: Value enclosed in single quotes
            '       # Match a single quote
            [^'\\]* # Match any character except quotes or backslashes zero or more times
            (?:\\.[^'\\]*)*  # Match any escaped character followed by any character except quotes or backslashes zero or more times
            '       # Match the closing quote
        )
        |           # Or
        ([\w\-.%:\/]+)  # Capture group 4: Value without quotes
    )
)?                  # End of optional non-capturing group for value
(?:$|\s)            # Match end of string or whitespace
""",
    re.VERBOSE,
)


class PostInitMeta(type):
    def __call__(cls, *args, **kwargs):
        obj = cast(Element, super().__call__(*args, **kwargs))
        obj._define_scope_id = get_current_scope().id
        get_current_container().add_child(obj)
        return obj


class Element(Renderable, metaclass=PostInitMeta):
    dependency: ClassVar[Optional[ComponentDependencyInfo]] = None

    def __init__(self, tag: Optional[Union[str, Variable]] = None):
        if self.dependency:
            tag = self.dependency.tag_name or ""

        super().__init__()

        self._tag = tag
        self._str_classes: list[str] = []
        self._map_classes: dict[str, Union[bool, Any]] = {}
        self._binding_classes: list[Any] = []
        self._style: dict[str, str] = {}
        self._binging_style: dict[str, Any] = {}
        self._proxy_style: list[Any] = []
        self._props: dict[str, Any] = {}
        self._binding_props: dict[str, Any] = {}
        self._proxy_props: list[Any] = []

        self._events: defaultdict[str, list[EventArgs]] = defaultdict(list)
        self._lifecycle_events: defaultdict[str, list[EventArgs]] = defaultdict(list)
        self._directives: dict[Directive, None] = {}
        self._slots: defaultdict[str, Slot] = defaultdict(Slot)
        self._element_ref: Optional[ElementRef] = None
        self._define_scope_id = sys.maxsize

        if self.dependency:
            add_component_dependency(type(self), self.dependency)
        trigger_element_create(self)

    def __init_subclass__(
        cls,
        *,
        esm: Union[str, Path, None] = None,
        externals: Optional[dict[str, Path]] = None,
        css: Union[list[Union[str, Path]], None] = None,
        zero_externals: Optional[dict[str, Path]] = None,
        zero_css: Union[list[Union[str, Path]], None] = None,
    ) -> None:
        super().__init_subclass__()

        if esm:
            esm = _make_dependency_path(esm, cls)

            if externals:
                externals = {
                    key: _make_dependency_path(value, cls)
                    for key, value in externals.items()
                }

            if css:
                css = set(_make_dependency_path(c, cls) for c in css)  # type: ignore

            tag_name = f"instaui-{esm.stem}"

            has_overrides = zero_externals or zero_css

            if not has_overrides:
                overrides = None
            else:
                overrides = {}

                overrides[RuntimeMode.ZERO] = ComponentDependencyOverride(
                    zero_externals=zero_externals, zero_css=zero_css
                )

            cls.dependency = ComponentDependencyInfo(
                tag_name=tag_name,
                esm=esm,
                externals=cast(dict[str, Path], externals or {}),
                css=cast(list[Path], css or list()),  # type: ignore
                overrides=overrides,
            )

    @property
    def slots(self):
        return self._slots

    @property
    def default_slot(self) -> Slot:
        return self._slots[DEFAULT_SLOT_NAME]

    def __enter__(self):
        self.default_slot.__enter__()
        return self

    def __exit__(self, *_):
        self.default_slot.__exit__(*_)

    def add_slot(self, name: str) -> Slot:
        return self._slots[name]

    @property
    def tag(self):
        assert self._tag is not None, "tag is not set"
        return self._tag

    def on_mounted(
        self,
        handler: EventMixin,
        *,
        params: Optional[list] = None,
    ):
        assert isinstance(handler, EventMixin), (
            "handler must be an instance of EventMixin"
        )

        handler._attach_to_element(params=params)
        args = EventArgs(MOUNTED_EVENT_NAME, handler, params=params)
        self._lifecycle_events[MOUNTED_EVENT_NAME].append(args)

        return self

    def scoped_style(
        self,
        style: str,
        *,
        selector: Union[str, Callable[[str], str]] = "*",
        with_self=False,
    ):
        """
        Applies scoped CSS styles to child elements within the component.

        Args:
            style (str): The CSS style rules to be applied within the scope.
            selector (Union[str, Callable[[str], str]], optional): CSS selector or function
                that generates a selector to target specific child elements. Defaults to "*".
            with_self (bool, optional): If True, applies the styles to the component itself
                in addition to its children. Defaults to False.

        Example:
        .. code-block:: python
            # Apply red outline to all direct children
            ui.column().scoped_style("outline: 1px solid red;")

            # Apply styles only to elements with specific class
            ui.box().scoped_style("color: blue;", selector=".target-class")

            # Apply styles to component itself and children
            ui.column().scoped_style("outline: 1px solid red;", with_self=True)

            # Use lambda function for dynamic selector generation
            ui.box().scoped_style(
                "outline: 1px solid red;",
                selector=lambda t: f"{t}:has(.hover) .target"
            )
        """
        app = get_app()
        ssid = app.gen_scoped_style_group_id()

        select_box = f"*[insta-scoped-style={ssid}]"
        real_selector = (
            f"{select_box} {selector}"
            if isinstance(selector, str)
            else selector(select_box)
        )

        if with_self:
            real_selector = f"{select_box},{real_selector}"

        real_selector = f":where({real_selector})"
        style_code = f"{real_selector} {{ {style} }}"

        self.props({"insta-scoped-style": ssid})
        assets.add_style_tag(style_code, group_id=SCOPED_STYLE_GROUP_ID)
        return self

    def slot_props(self, name: str):
        return self._slots[DEFAULT_SLOT_NAME].slot_props(name)

    @staticmethod
    def _update_classes(
        classes: list[str],
        add: str,
    ) -> list[str]:
        return list(dict.fromkeys(classes + add.split()))

    @staticmethod
    def _parse_style(text: Union[str, dict[str, str]]) -> dict[str, str]:
        if isinstance(text, dict):
            return text

        if not text:
            return {}

        result = {}
        for item in text.split(";"):
            item = item.strip()
            if item:
                key, value = item.split(":")
                key = key.strip()
                value = value.strip()
                result[key] = value

        return result

    @staticmethod
    def _parse_props(props: Union[str, dict[str, Any]]) -> dict[str, Any]:
        if isinstance(props, dict):
            return props

        if not props:
            return {}

        dictionary = {}
        for match in PROPS_PATTERN.finditer(props or ""):
            key = match.group(1)
            value = match.group(2) or match.group(3) or match.group(4)
            if value is None:
                dictionary[key] = True
            else:
                if (value.startswith("'") and value.endswith("'")) or (
                    value.startswith('"') and value.endswith('"')
                ):
                    value = ast.literal_eval(value)
                dictionary[key] = value
        return dictionary

    def key(self, key: Any):
        """Set the key prop of the component.

        Args:
            key (str): The key prop value.

        """
        self.props({"key": key})
        return self

    def vmodel(
        self,
        value: Any,
        modifiers: Union[TVModelModifier, list[TVModelModifier], None] = None,
        *,
        prop_name: str = "value",
        is_html_component=False,
    ):
        if prop_name == "value":
            prop_name = "modelValue"

        if isinstance(modifiers, str):
            modifiers = [modifiers]

        self.directive(
            Directive(
                is_sys=is_html_component,
                name="vmodel",
                arg=prop_name,
                modifiers=modifiers,
                value=value,  # type: ignore
            )
        )

        return self

    @overload
    def classes(self, add: str) -> Self: ...
    @overload
    def classes(self, add: dict[str, bool]) -> Self: ...

    @overload
    def classes(self, add: str) -> Self: ...

    def classes(
        self,
        add: Union[
            str,
            dict[str, bool],
            VForItem,
        ],
    ) -> Self:
        """
        Adds one or more CSS classes to the element, supporting static strings,
        reactive string references, or conditional class bindings.

        Args:
            add (str | dict[str, TMaybeRef[bool]] | TMaybeRef[str] | VForItem):
                CSS class configuration to apply. It can be:
                - A static class name string.
                - A reactive reference to a dynamic class name string.
                - A dictionary mapping class names to reactive boolean values,
                which toggle classes on or off.
                - A loop item reference when used inside v-for style bindings.

        Example:
        .. code-block:: python
            from instaui import ui, html

            # Static class
            html.span("target").classes("test")

            # Dynamic class string
            ref = ui.state("c1")
            html.span("target").classes(ref)

            # Conditional class binding
            c1 = ui.state(True)
            c2 = ui.state(False)
            html.span("target").classes({"c1": c1, "c2": c2})
        """

        mark_used(add, host_scope_id=self._define_scope_id)

        if isinstance(add, str):
            self._str_classes = self._update_classes(self._str_classes, add)

        elif isinstance(add, dict):
            self._map_classes.update(add)

        elif is_bindable(add):
            self._binding_classes.append(add)

        return self

    def style(self, add: Union[str, dict[str, Any]]) -> Self:
        """
        Applies inline CSS styles to the element. Supports static strings, dictionaries, or reactive references
        to dynamically update styles.

        Args:
            add (Union[str, dict[str, Any]]): The style(s) to apply. Can be:
                - A CSS string (e.g., "color: red;")
                - A dictionary mapping CSS properties to values (e.g., {"color": "red"})
                - A reactive reference to a string or dictionary for dynamic updates.

        Example:
        .. code-block:: python
            from instaui import ui, html

            # Apply static string style
            html.span("inline style").style("color: red;")

            # Apply dictionary style
            ref = ui.state("red")
            html.span("target").style({"color": ref})

            # Apply reactive string style
            style = ui.state("color: red;")
            html.span("target").style(style)
        """

        mark_used(add, host_scope_id=self._define_scope_id)

        if isinstance(add, str):
            self._style.update(self._parse_style(add))

        elif isinstance(add, dict):
            self._style.update(
                {key: value for key, value in add.items() if not is_bindable(value)}
            )
            self._binging_style.update(
                {key: value for key, value in add.items() if is_bindable(value)}
            )

        elif is_bindable(add):
            self._proxy_style.append(add)
            return self

        return self

    def props(self, add: Union[str, dict[str, Any]]) -> Self:
        """
        Applies one or more HTML properties to the element. Supports constant values,
        string boolean attributes, reactive bindings, and dynamic evaluated props.

        Args:
            add (Union[str, dict[str, Any]]):
                The property source to apply.
                - If a string, the property is treated as a boolean attribute (e.g., "disabled").
                - If a dict, key-value pairs are applied as element properties.
                - If a reactive reference (state), the property updates automatically
                when the referenced value changes.

        Example:
        .. code-block:: python
            from instaui import ui, html

            # Apply constant dictionary
            html.button("Submit").props({"disabled": True})

            # Apply boolean attribute using a string
            html.button("Click").props("disabled")

            # Bind reactive state to a property
            value = ui.state(True)
            html.checkbox(value)
            html.button("Submit").props({"disabled": value})

            # Apply dictionary state with multiple properties
            value = ui.state({"disabled": True})
            html.button("target").props(value)
        """

        mark_used(add, host_scope_id=self._define_scope_id)

        if isinstance(add, str):
            self._props.update(self._parse_props(add))

        elif is_bindable(add):
            self._proxy_props.append(add)
            return self

        elif isinstance(add, dict):
            self._props.update(
                {
                    key: value
                    for key, value in add.items()
                    if (not is_bindable(value)) and value is not None
                }
            )

            self._binding_props.update(
                {key: value for key, value in add.items() if is_bindable(value)}
            )

        return self

    def on(
        self,
        event_name: str,
        handler: EventMixin,
        *,
        params: Optional[list] = None,
        modifier: Optional[list[TEventModifier]] = None,
    ):
        """
        Attaches an event handler to the element. Supports modifiers for event handling,

        Args:
            event_name (str): The name of the event to listen for.
            handler (EventMixin): The event handler to attach.
            params (Optional[list], optional): A list of values corresponding to handler parameters defined with `ui.event_param()`.
                The order of values in the list will match the order of the placeholders in the
                handler function signature. This allows dynamic context-specific values to be
                injected into the event handler at runtime.
            modifier (Optional[list[TEventModifier]], optional): A list of modifiers to apply. Defaults to None.

        """

        assert isinstance(handler, EventMixin), (
            "handler must be an instance of EventMixin"
        )
        mark_used(handler, host_scope_id=self._define_scope_id)

        handler._attach_to_element(params=params, modifier=modifier)
        event_name, modifier = _parse_event_modifiers(event_name, modifier)
        args = EventArgs(event_name, handler, params=params, modifier=modifier)
        self._events[event_name].append(args)

        return self

    def directive(self, directive: Directive) -> Self:
        self._directives[directive] = None
        return self

    def display(self, value: bool) -> Self:
        return self.directive(Directive(is_sys=False, name="vshow", value=value))

    def event_dataset(self, data: Any, name: str = "event-data") -> Self:
        name = f"data-{name}"
        self.props({name: to_json_str(data)})
        return self

    def element_ref(self, ref: ElementRef):
        """
        Associates an `ElementRef` with the component, allowing interaction with
        the underlying UI element from Python or through event callbacks.

        Args:
            ref (ElementRef): A reference object used to access or manipulate
                the rendered UI element programmatically.

        Example:
        .. code-block:: python
            from instaui import ui, html
            from custom import Counter

            cp = ui.element_ref()

            @ui.event(outputs=[cp])
            def on_click():
                return ui.run_element_method("reset")

            Counter().element_ref(cp)
            html.button("reset").on_click(on_click)
        """

        mark_used(ref, host_scope_id=self._define_scope_id)
        self._element_ref = ref
        return self

    def use(self, *use_fns: Callable[[Self], None]) -> Self:
        """Use functions to the component object.

        Args:
            use_fns (Callable[[Self], None]): The list of use functions.

        Examples:
        .. code-block:: python
            def use_red_color(element: html.paragraph):
                element.style('color: red')

            html.paragraph('Hello').use(use_red_color)
        """

        for fn in use_fns:
            fn(self)
        return self

    @classmethod
    def use_init(cls, init_fn: Callable[[type[Self]], Self]) -> Self:
        """Use this method to initialize the component.

        Args:
            init_fn (Callable[[type[Self]], Self]): The initialization function.

        Examples:
        .. code-block:: python
            def fack_init(cls: type[html.table]) -> html.table:
                return cls(columns=['name', 'age'],rows = [{'name': 'Alice', 'age': 25}, {'name': 'Bob', 'age': 30}])

            ui.table.use_init(fack_init)
        """
        return init_fn(cls)


def _make_dependency_path(path: Union[str, Path], cls: type):
    if isinstance(path, str):
        path = Path(path)

    if not path.is_absolute():
        path = Path(inspect.getfile(cls)).parent / path

    return path


def _parse_event_modifiers(
    event_name: str,
    org_modifier: Optional[list[TEventModifier]] = None,
) -> tuple[str, Optional[list[TEventModifier]]]:
    """Parse event name and modifiers from both event_name string and org_modifier list.

    Args:
        event_name: Event name string, may contain modifiers separated by dots (e.g. 'click.stop')
        org_modifier: Optional list of additional modifiers

    Returns:
        Tuple of (cleaned_event_name, combined_modifiers) where:
            - cleaned_event_name: event name without modifiers
            - combined_modifiers: tuple of unique modifiers from both sources
    """
    parts = event_name.split(".")
    base_name = parts[0]
    modifiers = [m.strip() for m in parts[1:]] if len(parts) > 1 else []

    if not org_modifier and not modifiers:
        return base_name, None

    combined = set(modifiers)
    if org_modifier:
        combined.update(org_modifier)

    return base_name, cast(list[TEventModifier], list(combined)) if combined else None



# from typing import Any, Optional
# from instaui.systems.dataclass_system import dataclass


# @dataclass(frozen=True)
# class ScriptTag:
#     content: str
#     script_attrs: Optional[dict[str, Any]] = None


from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Literal


ScriptLoading = Literal["async", "defer", "blocking"]
ScriptPosition = Literal["head", "body"]


@dataclass(slots=True, frozen=True)
class JSAsset:
    kind: Literal["inline", "file", "url"]
    source: str | Path
    module: bool = False
    loading: ScriptLoading = "blocking"
    position: ScriptPosition = "body"
    attrs: dict[str, Any] = field(default_factory=dict)



from __future__ import annotations
from contextlib import contextmanager
import contextvars
from pathlib import Path
from typing import TYPE_CHECKING, Any, Hashable, Optional, Union
from instaui.internal.assets.component_dep import ComponentDependencyInfo
from instaui.internal.assets.css_assets import CssAsset, CssRole
from instaui.internal.assets.plugin import PluginDependencyInfo, PluginOptionsBuilder
from instaui.internal.assets.script_assets import (
    JSAsset,
    ScriptLoading,
    ScriptPosition,
)
from instaui.internal.assets.style_assets import StyleTag
from instaui.internal.context import in_page_context
from .base import AssetsDeclaration

if TYPE_CHECKING:
    from instaui.internal.ui.element import Element

GLOBAL_ASSETS = AssetsDeclaration()
LOCAL_ASSETS: contextvars.ContextVar[Optional[AssetsDeclaration]] = (
    contextvars.ContextVar("local_assets", default=None)
)


@contextmanager
def enter_assets_context():
    token = LOCAL_ASSETS.set(AssetsDeclaration())
    try:
        yield
    finally:
        LOCAL_ASSETS.reset(token)


def get_active_assets() -> AssetsDeclaration:
    """
    Return the AssetsPool that should actually be written based on context.
    """
    if in_page_context():
        local = LOCAL_ASSETS.get()
        if local is None:
            local = AssetsDeclaration()
            LOCAL_ASSETS.set(local)
        return local
    else:
        return GLOBAL_ASSETS


def add_css_link(
    css: Union[Path, str],
    *,
    role: Union[str, CssRole] = CssRole.COMPONENT,
    namespace: Optional[Hashable] = None,
):
    """Add a CSS link to the HTML document."""
    role = CssRole(role) if isinstance(role, str) else role
    get_active_assets().css_links.add(css, namespace=namespace, role=role)


def remove_css_link(
    css: Union[Path, str, CssAsset],
    *,
    role: Union[str, CssRole] = CssRole.COMPONENT,
    namespace: Optional[Hashable] = None,
):
    role = CssRole(role) if isinstance(role, str) else role
    get_active_assets().css_links.remove(css, namespace=namespace, role=role)


def use_favicon(path: Path):
    """Set the favicon of the HTML document.

    Args:
        favicon (Path): The path to the favicon.
    """
    get_active_assets().favicon = path


def add_import_map(name: str, link: Union[str, Path]):
    """Add an import map to the HTML document."""
    get_active_assets().import_maps[name] = link


def register_plugin(
    name: str,
    esm: Path,
    *,
    externals: Optional[dict[str, Path]] = None,
    css: Optional[list[Union[str, Path, CssAsset]]] = None,
    options: Optional[dict | PluginOptionsBuilder] = None,
):
    """
    Register a plugin

    Args:
        name (str): Plugin name
        esm (Path): Path to plugin's ESM file. Format should match vue.js plugins
        externals (Optional[dict[str, Path]], optional): External dependencies for plugin JS implementation. Defaults to None.
        css (Optional[list[Path]], optional): List of CSS file paths for the plugin. Defaults to None.
        options (Optional[dict | PluginOptionsBuilder], optional): Plugin options. Defaults to None.

    Example:
    .. code-block:: python
        from instaui import custom

        # Basic usage
        custom.register_plugin("my-plugin", Path(__file__).parent / "my-plugin.esm.js")

        # With external dependencies
        custom.register_plugin(
            "my-plugin",
            Path(__file__).parent / "my-plugin.esm.js",
            externals={"echarts": Path(__file__).parent / "echarts.esm.js"})

        # With install method options
        custom.register_plugin(..., options = {'opt1': 'value1', 'opt2': 'value2'})

        # With component-dependent options (only needed for complex plugins that define components, like vue-router)
        from instaui import custom

        def parse_scope(self, parse_key: StringKeyFn, parse_scope: ParseScopeFn):
            # parse_key ensures JS object keys are properly quoted
            # parse_scope translates Scope objects into component references
            return {parse_key(k): parse_scope(v) for k, v in self.routes.items()}

        custom.register_plugin(..., options = parse_scope)
    """
    info = PluginDependencyInfo(
        f"plugin/{name}", esm, externals or {}, css or [], options=options
    )
    get_active_assets().plugins.add(info)
    return info


def clear_registered_plugins():
    get_active_assets().plugins.clear()


def add_js_inline(
    code: str,
    *,
    module: bool = False,
    position: ScriptPosition = "body",
    attrs: Optional[dict[str, Any]] = None,
) -> None:
    """
    Add inline JavaScript code to the document.

    Args:
        code: JavaScript source code.
        module: Whether to emit <script type="module">.
        position: Where to inject the script: "head" or "body".
        attrs: Extra attributes for the <script> tag (e.g. nonce, crossorigin, data-*).

    Example:
        ui.add_js_inline("console.log('hi')", module=True)
    """
    if not isinstance(code, str) or not code.strip():
        raise ValueError("code must be a non-empty string")

    asset = JSAsset(
        kind="inline",
        source=code,
        module=module,
        position=position,
        attrs=attrs or {},
    )

    get_active_assets().js_asset.append(asset)


def add_js_file(
    path: Path | str,
    *,
    module: bool = False,
    loading: ScriptLoading = "blocking",
    position: ScriptPosition = "body",
    attrs: Optional[dict[str, Any]] = None,
) -> None:
    """
    Add a local JavaScript file as an asset.

    The file will be handled by the runtime (e.g. registered as a static file in web mode).

    Args:
        path: Path to the JS file.
        module: Whether to emit <script type="module">.
        loading: Script loading behavior.
        position: Where to inject the script tag.
        attrs: Extra script tag attributes.

    Example:
        ui.add_js_file(Path("assets/app.js"), module=True)
    """
    p = Path(path)

    if not p.exists():
        raise FileNotFoundError(p)

    if not p.is_file():
        raise ValueError(f"Not a file: {p}")

    asset = JSAsset(
        kind="file",
        source=p,
        module=module,
        loading=loading,
        position=position,
        attrs=attrs or {},
    )

    get_active_assets().js_asset.append(asset)


def remove_js_file(path: Path | str):
    p = Path(path)

    if not p.exists():
        return

    if not p.is_file():
        raise ValueError(f"Not a file: {p}")

    target = p.resolve()
    assets = get_active_assets().js_asset

    for i in range(len(assets) - 1, -1, -1):
        asset = assets[i]

        if asset.kind != "file":
            continue

        try:
            src = Path(asset.source).resolve()
        except Exception:
            continue

        if src == target:
            del assets[i]


def add_js_url(
    url: str,
    *,
    module: bool = False,
    loading: ScriptLoading = "blocking",
    position: ScriptPosition = "body",
    attrs: Optional[dict[str, Any]] = None,
) -> None:
    """
    Add a remote JavaScript file via URL.

    Args:
        url: Absolute or protocol-relative URL.
        module: Whether to emit <script type="module">.
        loading: Script loading behavior.
        position: Where to inject the script tag.
        attrs: Extra script tag attributes.

    Example:
        ui.add_js_url("https://cdn.jsdelivr.net/npm/vue@3/dist/vue.global.js")
    """
    if not isinstance(url, str) or not url.strip():
        raise ValueError("url must be a non-empty string")

    asset = JSAsset(
        kind="url",
        source=url,
        module=module,
        loading=loading,
        position=position,
        attrs=attrs or {},
    )

    get_active_assets().js_asset.append(asset)


def add_style_tag(
    content: str, *, attrs: dict | None = None, group_id: str | None = None
):
    """Add a style tag to the HTML document.

    Args:
        content (str): The content of the style tag.
        attrs: Tag attributes.
        group_id (Optional[str], optional): The group id of the style tag. Defaults to None.
    """
    get_active_assets().style_tags.append(StyleTag(content, group_id, attrs))


def add_component_dependency(element: type[Element], info: ComponentDependencyInfo):
    get_active_assets().component_dependencies.add(element, info)
    return info


def register_component_extension(
    *,
    target: type[Element],
    kind: str,
    values: list[str],
):
    """
    Invoked by third-party component authors to declare the additional dependencies required for a component family in Zero mode.
    """
    get_active_assets().component_extensions.add(
        target=target,
        kind=kind,
        values=values,
    )



from typing import ClassVar
from typing_extensions import Self
from contextvars import ContextVar


class PageState:
    """
    Returns the current PageState instance for this subclass, creating one if it does not yet exist.
    This ensures each page request has an isolated state container via a ContextVar.

    Args:
        cls (type): The subclass of PageState requesting its context-bound instance.

    Returns:
        Self: The existing or newly created PageState instance associated with the current context.

    Example:
    .. code-block:: python
        class MyState(PageState):
            def __init__(self):
                self.a1 = ui.state("foo")

        # Must call prepare() at the beginning of the page function
        MyState.prepare()

        # Then you can get the state instance anywhere
        my_state = MyState.get()
        value = my_state.a1
    """

    _ctx: ClassVar[ContextVar["PageState"]]
    _prepared: ClassVar[bool] = False

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._ctx = ContextVar(f"{cls.__name__}_ctx")
        cls._prepared = False

    @classmethod
    def prepare(cls) -> Self:
        """
        Prepare the page state instance. Must be called at the beginning of the page function,
        outside of any component scope.

        Returns:
            Self: The newly created PageState instance.

        Raises:
            RuntimeError: If prepare() is called multiple times.
        """
        if cls._prepared:
            raise RuntimeError(
                f"{cls.__name__}.prepare() has already been called. "
                "Do not call prepare() multiple times in the same page."
            )

        inst = cls()
        cls._ctx.set(inst)
        cls._prepared = True
        return inst  # type: ignore

    @classmethod
    def get(cls) -> Self:
        """
        Get the page state instance for the current context.

        Returns:
            Self: The page state instance.

        Raises:
            RuntimeError: If get() is called before prepare().
        """
        if not cls._prepared:
            raise RuntimeError(
                f"Must call {cls.__name__}.prepare() at the beginning of the page function "
                f"before calling {cls.__name__}.get()"
            )

        inst = cls._ctx.get(None)

        # This line should never execute since _prepared guarantees instance exists
        # But kept for robustness
        if inst is None:
            raise RuntimeError(
                f"Internal error: {cls.__name__}._prepared is True but no instance found in context"
            )

        return inst  # type: ignore



import asyncio
import threading
from typing import Any, Coroutine
import weakref
import contextlib
import atexit


class AsyncExecutor:
    """
    Thread-safe executor for running asyncio coroutines in any thread.

    Each thread gets its own event loop, managed independently.
    """

    def __init__(self):
        self._local = threading.local()
        self._cleaners = []
        self._lock = threading.Lock()

        atexit.register(self._cleanup_all)

    def _get_loop(self) -> asyncio.AbstractEventLoop:
        """Get or create an event loop for the current thread"""
        if not hasattr(self._local, "loop") or self._local.loop.is_closed():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            self._local.loop = loop

            # Register automatic cleanup when thread exits
            self._register_thread_cleanup(loop)

        return self._local.loop

    def _register_thread_cleanup(self, loop: asyncio.AbstractEventLoop):
        """Register cleanup to run when thread exits"""

        def cleanup():
            with contextlib.suppress(RuntimeError, Exception):
                if not loop.is_closed():
                    # Cancel all tasks
                    for task in asyncio.all_tasks(loop):
                        task.cancel()

                    # Run loop briefly to let tasks cancel
                    if not loop.is_running() and not loop.is_closed():
                        try:
                            loop.run_until_complete(asyncio.sleep(0))
                        except (RuntimeError, asyncio.CancelledError):
                            pass

                    loop.close()

        thread = threading.current_thread()
        finalizer = weakref.finalize(thread, cleanup)

        with self._lock:
            self._cleaners = [f for f in self._cleaners if f.alive]
            self._cleaners.append(finalizer)

    def run(self, coro: Coroutine) -> Any:
        """
        Run coroutine in the current thread's event loop.

        Args:
            coro: Coroutine to execute

        Returns:
            Coroutine result

        Raises:
            RuntimeError: If event loop is in invalid state
        """
        loop = self._get_loop()

        if loop.is_closed():
            raise RuntimeError("Event loop is closed")

        # If loop is already running, use thread-safe execution
        if loop.is_running():
            future = asyncio.run_coroutine_threadsafe(coro, loop)
            try:
                return future.result()
            finally:
                pass
        else:
            # Run in current thread
            try:
                return loop.run_until_complete(coro)
            finally:
                # Don't close the loop here - it might be reused
                pass

    def clear_current_thread(self):
        """Clean up the event loop for the current thread"""
        if hasattr(self._local, "loop"):
            loop = self._local.loop
            if not loop.is_closed():
                # Cancel all tasks
                for task in asyncio.all_tasks(loop):
                    task.cancel()

                # Let tasks cancel
                if not loop.is_running():
                    with contextlib.suppress(asyncio.CancelledError, RuntimeError):
                        loop.run_until_complete(asyncio.sleep(0))

                loop.close()

            delattr(self._local, "loop")

    def _cleanup_all(self):
        """Clean up all loops (called at program exit)"""
        if hasattr(self._local, "loop"):
            self.clear_current_thread()

        with self._lock:
            for finalizer in self._cleaners:
                if finalizer.alive:
                    finalizer()
            self._cleaners.clear()



from typing import Callable, Mapping, Optional
import pydantic_core
from instaui.internal.codec.adapter import empty_type_adapter
from instaui.protocol.codec.type_adapter import TypeAdapterProtocol
from instaui.systems.dataclass_system import dataclass, field
from instaui.systems import func_system


@dataclass()
class HandlerInfo:
    fn: Callable
    fn_location_info: str
    outputs_binding_count: int = 0
    handler_param_converters: list[TypeAdapterProtocol] = field(default_factory=list)
    is_last_param_args: bool = False

    def get_handler_args(self, input_values: list):
        real_param_converters = _try_expand_params_converters(
            self.handler_param_converters, input_values, self.is_last_param_args
        )

        try:
            return [
                param_converter(value)
                for param_converter, value in zip(real_param_converters, input_values)
            ]
        except pydantic_core._pydantic_core.ValidationError as e:
            raise ValueError(f"invalid input[{self.fn_location_info}]: {e}") from None

    @classmethod
    def from_handler(
        cls,
        handler: Callable,
        outputs_binding_count: int,
        custom_type_adapter_map: Optional[Mapping[int, TypeAdapterProtocol]] = None,
        *,
        skip_convert_param=False,
    ):
        custom_type_adapter_map = custom_type_adapter_map or {}
        params_infos = func_system.get_fn_params_infos(handler)
        is_last_param_args = func_system.is_last_param_args(handler)
        param_converters = (
            []
            if skip_convert_param
            else [
                custom_type_adapter_map.get(idx, empty_type_adapter)
                for idx, (_, param_type) in enumerate(params_infos)
            ]
        )

        file, lineno, _ = func_system.get_function_location_info(handler)

        return cls(
            handler,
            f'File "{file}", line {lineno}',
            outputs_binding_count,
            handler_param_converters=param_converters,
            is_last_param_args=is_last_param_args,
        )


def _try_expand_params_converters(
    old_param_converters: list[TypeAdapterProtocol],
    input_values: list,
    is_last_param_args: bool,
):
    if not is_last_param_args:
        return old_param_converters

    diff = len(input_values) - len(old_param_converters)
    if diff == 0:
        return old_param_converters

    arg_param_converters = [old_param_converters[-1]] * diff

    return [*old_param_converters[:], *arg_param_converters]



from instaui.internal.ui.element import Element
from instaui.internal.ui.value_element import ValueElement
from instaui.internal.ui.bindable import BindableMixin
from instaui.types import TVModelModifier
from ._mixins import InputEventMixin

_T_value = int | float


class Range(InputEventMixin, ValueElement[_T_value]):
    def __init__(
        self,
        value: _T_value | None = None,
        *,
        min: _T_value | None = None,
        max: _T_value | None = None,
        step: _T_value | None = None,
        **kwargs,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props({"type": "range", "min": min, "max": max, "step": step, **kwargs})

    def vmodel(
        self,
        value: BindableMixin,
        modifiers: TVModelModifier | list[TVModelModifier] | None = None,
        *,
        prop_name: str = "value",
    ):
        modifiers = modifiers or []
        if isinstance(modifiers, str):
            modifiers = [modifiers]

        modifiers_with_number = list(set([*modifiers, "number"]))

        return super().vmodel(value, modifiers_with_number, prop_name=prop_name)  # type: ignore

    def _input_event_mixin_element(self) -> Element:
        return self



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Optional
from dataclasses import dataclass, field
from jinja2 import Environment, FileSystemLoader
from instaui.internal.codegen.program_codegen import CodegenResult
from .html_context import HtmlRenderContext

if TYPE_CHECKING:
    from instaui.internal.assets.style_assets import StyleTag


class WebViewHtmlRenderer:
    def __init__(self):
        self.env = Environment(
            loader=FileSystemLoader(Path(__file__).parent / "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def render(self, codegen: CodegenResult, options: HtmlOptions) -> str:
        template = self.env.get_template("webview.html.j2")

        ctx = HtmlRenderContext(
            app_var=codegen.app_var,
            title=options.title,
            favicon_url=options.favicon_url,
            css_links=options.css_links,
            style_tags=[
                (tag.group_id, tag.content, tag.attrs or {})
                for tag in options.style_tags
            ],
            script_tags_in_head=options.script_tags_in_head,
            script_tags_in_body=options.script_tags_in_body,
            entry_code=codegen.entry_code,
            import_map=options.import_map if options.import_map else None,
            icons_svg_code=options.icons_svg_code,
        )

        return template.render(model=ctx)


@dataclass()
class HtmlOptions:
    favicon_url: str
    title: str = "instaui"

    css_links: list[str] = field(default_factory=list)
    style_tags: list[StyleTag] = field(default_factory=list)
    script_tags_in_head: list[tuple[Optional[dict], str | None]] = field(
        default_factory=list
    )
    """(attrs_dict, content)"""
    script_tags_in_body: list[tuple[Optional[dict], str | None]] = field(
        default_factory=list
    )
    """(attrs_dict, content)"""
    import_map: Optional[dict[str, str]] = None
    icons_svg_code: Optional[str] = None
